Import "Vector3.bmx"
Import "Matrix4.bmx"

Const Q_EPS# = 0.0001'1

  Type Quaternion
    field x#=0,y#=0,z#=0,w#=1
    Method this:Float Ptr()
      return varptr(x)
    EndMethod

    Method Zero()
      x=0.0
      y=0.0
      z=0.0
    EndMethod
    Method Set(tx#,ty#,tz#,tw#)
      x=tx
      y=ty
      z=tz
      w=tw
    EndMethod
    Method Copy( q:Quaternion )
        x = q.x
        y = q.y
        z = q.z
        w = q.w
    EndMethod
    Method Clone( p:Quaternion var )
      p.x = x
      p.y = y
      p.z = z
      p.w = w
    EndMethod


    Method SetEuler( xx#,yy#,zz# )
      local sp#=sin(-yy/2.0), cp#=cos(-yy/2.0)
	    local sy#=sin(zz/2.0), cy#=cos(zz/2.0)
	    local sr#=sin(xx/2.0), cr#=cos(xx/2.0)

	    w=cr*cp*cy + sr*sp*sy;
	    x=sr*cp*cy - cr*sp*sy;
	    y=cr*sp*cy + sr*cp*sy;
	    z=cr*cp*sy - sr*sp*cy;

    EndMethod

    Method GetMatrix(ans:Matrix4 var)
      ans.Identity()
      local xx#=x*x
      local yy#=y*y
      local zz#=z*z
    	local xy#=x*y
      local xz#=x*z
      local yz#=y*z
      local wx#=w*x
      local wy#=w*y
      local wz#=w*z

      ans.m[0] =1-2*(yy+zz)
      ans.m[1] =  2*(xy-wz)
      ans.m[2] =  2*(xz+wy)
    	ans.m[4] =  2*(xy+wz)
      ans.m[5] =1-2*(xx+zz)
      ans.m[6] =  2*(yz-wx)
    	ans.m[8] =  2*(xz-wy)
      ans.m[9] =  2*(yz+wx)
      ans.m[10]=1-2*(xx+yy)

    EndMethod

    Method MagnitudeSqr#()
      return (x*x)+(y*y)+(z*z)+(w*w)
    EndMethod
    Method Magnitude#()
      return sqr((x*x)+(y*y)+(z*z)+(w*w))
    EndMethod
    Method Add(ans:Quaternion var, q2:Quaternion)
      ans.x=x + q2.x
      ans.y=y + q2.y
      ans.z=z + q2.z
      ans.w=w + q2.w
    EndMethod
    Method Sub(ans:Quaternion var, q2:Quaternion)
      ans.x=x - q2.x
      ans.y=y - q2.y
      ans.z=z - q2.z
      ans.w=w - q2.w
    EndMethod
    Method Multiply(ans:Quaternion var, q2:Quaternion)
      local cx#,cy#,cz#,cw#
    	cw# = (w*q2.w) - (x*q2.x) - (y*q2.y) - (z*q2.z)
    	cx# = (w*q2.x) + (x*q2.w) + (y*q2.z) - (z*q2.y)
    	cy# = (w*q2.y) + (y*q2.w) + (z*q2.x) - (x*q2.z)
    	cz# = (w*q2.z) + (z*q2.w) + (x*q2.y) - (y*q2.x)
      ans.Set(cx,cy,cz,cw)
    EndMethod

    Method Invert( ans:Quaternion var )
      Local l#
	    l = sqr((x*x)+(y*y)+(z*z)+(w*w))
    	If (l <> 0)
        l = 1.0 / l
    		ans.w = w*l
		    ans.x = x*l
    		ans.y = y*l
		    ans.z = z*l
    	EndIf
    EndMethod

    Method Conjugate(ans:Quaternion var)
      ans.w =  w
      ans.x = -x
      ans.y = -y
      ans.z = -z
    EndMethod

    Method Dot#(q2:Quaternion)
      local q#=0
    	q=(w*q2.w) + (x*q2.x) + (y*q2.y) + (z*q2.z)
    	Return q
    EndMethod

    Method Slerp(ans:Quaternion var, fin:Quaternion ,t#)
      Local scaler_w#, scaler_x#, scaler_y#, scaler_z#
      Local omega#, cosom#, sinom#, scale0#, scale1#

      cosom = Dot(fin)
      If cosom <= 0.0
        cosom = -cosom
    		scaler_w = -fin.w
    		scaler_x = -fin.x
    		scaler_y = -fin.y
    		scaler_z = -fin.z
    	Else
    		scaler_w = fin.w
    		scaler_x = fin.x
    		scaler_y = fin.y
    		scaler_z = fin.z
      EndIf

      If (1 - cosom) > Q_EPS
        omega = ACos(cosom)
        sinom = Sin(omega)
        scale0 = Sin((1 - t) * omega) / sinom
        scale1 = Sin(t * omega) / sinom
    	Else
        scale0 = 1 - t
        scale1 = t
    	EndIf
      ans.x = ( scale0 * x + scale1 * scaler_x )
      ans.y = ( scale0 * y + scale1 * scaler_y )
      ans.z = ( scale0 * z + scale1 * scaler_z )
      ans.w = ( scale0 * w + scale1 * scaler_w )
    EndMethod
    Method GetForwardAxis(ans:Vector3 var)
    	Local x2# = x + x
    	Local y2# = y + y
    	Local z2# = z + z
    	Local xx# = x * x2
    	Local xz# = x * z2
    	Local yy# = y * y2
    	Local yz# = y * z2
    	Local wx# = w * x2
    	Local wy# = w * y2
    	ans.x = -(xz+wy)
      ans.y = -(yz-wx)
      ans.z = -(1-(xx+yy))
    EndMethod


    Function Create:Quaternion(x#, y#, z#, w#)
      local q:Quaternion = new Quaternion
        q.x = x
        q.y = y
        q.z = z
        q.w = w
      return q
    EndFunction
  End Type               

